from setuptools import setup, find_packages

setup(
    name='EthicML',
    version='0.1.0-alpha',
    author='O. Thomas',
    author_email="olliethomas86@gmail.com",
    description="A toolkit for understanding and researching algorithmic bias",
    url="https://github.com/predictive-analytics-lab/EthicML",
    packages=find_packages(),
    package_data={'ethicml.data': ['csvs/*.csv']},
    python_requires=">=3.6",
    install_requires=[
        "imageio >= 2.4.1",
        "matplotlib >= 3.0.2",
        "numpy >= 1.14.2",
        "pandas >= 0.24.0",
        "scikit_learn >= 0.20.1",
        "seaborn >= 0.9.0",
        "torch >= 1.1.0",
        "pyarrow >= 0.11",
        "numba",
        "fairlearn >= 0.2.0",
        "GitPython >= 2.1.11",
        "tqdm >= 4.31.1",
        "pipenv >= 2018.11.26",
        "dataclasses;python_version<\"3.7\"",  # dataclasses are in the stdlib in python>=3.7
    ],
    extras_require={
        'dev': [  # use `pip install ethicml[dev]` to install development packages
            "pylint >= 2.0",
            "pytest >= 3.3.2",
            "pytest-cov >= 2.6.0",
            "mypy >= 0.710",
            "black",
        ],
    },
)



"""
Wrapper for calling Kamishima model
"""
from pathlib import Path
from tempfile import TemporaryDirectory

import numpy as np
import pandas as pd

from ethicml.algorithms.inprocess.installed_model import InstalledModel


class Kamishima(InstalledModel):
    """
    Model that calls Kamishima's code. Based on Algo-Fairness
    https://github.com/algofairness/fairness-comparison/blob/master/fairness/algorithms/kamishima/KamishimaAlgorithm.py
    """

    def __init__(self, eta=1.0):
        super().__init__(
            name="kamishima",
            url="https://github.com/predictive-analytics-lab/kamfadm.git",
            module="kamfadm",
            file_name="train_pr.py",
        )
        self.eta = eta

    @staticmethod
    def create_file_in_kamishima_format(data, file_path):
        """Create a text file with the data"""

        result = pd.concat([data.x, data.s, data.y], axis="columns").to_numpy().astype(np.float64)
        np.savetxt(file_path, result)

    async def run_async(self, train, test):
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_path = str(tmp_path / "train.txt")
            test_path = str(tmp_path / "test.txt")
            self.create_file_in_kamishima_format(train, train_path)
            self.create_file_in_kamishima_format(test, test_path)
            min_class_label = train.y[train.y.columns[0]].min()
            model_path = str(tmp_path / "model")
            output_path = str(tmp_path / "output.txt")

            try:
                await self._call_script(
                    [
                        str(self._module_path() / "train_pr.py"),
                        "-e",
                        str(self.eta),
                        "-i",
                        train_path,
                        "-o",
                        model_path,
                        "--quiet",
                    ]
                )

                await self._call_script(
                    [
                        str(self._module_path() / "predict_lr.py"),
                        "-i",
                        test_path,
                        "-m",
                        model_path,
                        "-o",
                        output_path,
                        "--quiet",
                    ]
                )
                output = np.loadtxt(output_path)
                predictions = output[:, 1].astype(np.float32)
            except RuntimeError:
                predictions = np.ones_like(test.y.to_numpy())

        to_return = pd.DataFrame(predictions, columns=["preds"])

        if to_return["preds"].min() != to_return["preds"].max():
            to_return = to_return.replace(to_return["preds"].min(), min_class_label)
        return to_return

    @property
    def name(self):
        return "Kamishima"



"""
Method for calculating the HSIC - an independence criterion. a score of 0 denotes independence
"""
import math
import pandas as pd
import numpy as np

from ethicml.utility.data_structures import DataTuple
from ethicml.metrics.metric import Metric


def hsic(
    prediction: np.ndarray, label: np.ndarray, sigma_first: float, sigma_second: float
) -> float:
    """Calculate the HSIC value"""
    xx_gram = np.array(np.matmul(np.expand_dims(prediction, 1), np.expand_dims(prediction, 1).T))
    yy_gram = np.array(np.matmul(np.expand_dims(label, 1), np.expand_dims(label, 1).T))

    x_sqnorms = np.diag(xx_gram)
    y_sqnorms = np.diag(yy_gram)

    exp_r = lambda x: np.expand_dims(x, 0)
    exp_c = lambda x: np.expand_dims(x, 1)

    gamma_first = 1.0 / (2 * sigma_first ** 2)
    gamma_second = 1.0 / (2 * sigma_second ** 2)
    # use the second binomial formula
    kernel_xx = np.exp(-gamma_first * (-2 * xx_gram + exp_c(x_sqnorms) + exp_r(x_sqnorms)))
    kernel_yy = np.exp(-gamma_second * (-2 * yy_gram + exp_c(y_sqnorms) + exp_r(y_sqnorms)))

    kernel_xx_mean = np.mean(kernel_xx)
    kernel_yy_mean = np.mean(kernel_yy)

    h_k = kernel_xx - kernel_xx_mean
    h_l = kernel_yy - kernel_yy_mean

    num = float(kernel_yy.shape[0])
    h_kf = h_k / (num - 1)
    h_lf = h_l / (num - 1)

    # biased estimate
    hsic_value = float(np.trace(np.matmul(h_kf.T, h_lf)))
    return hsic_value


class Hsic(Metric):
    """
    see module string
    """

    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        """
        We add the ability to take the average of hsic sscore as for
        larger datasets it will kill your machine
        """

        s_cols = actual.s.columns
        sens_labels = np.array(actual.s[s_cols].values)

        batchs_size = 5000

        together = np.hstack((prediction.values, sens_labels)).transpose()

        np.random.seed(888)
        col_idx = np.random.permutation(together.shape[1])

        together = together[:, col_idx]

        prediction_shuffled = together[0]
        label_shuffled = together[1]

        num_batches_float = prediction.shape[0] / batchs_size
        num_batches: int = int(math.ceil(num_batches_float))

        batches = []

        start = 0

        for _ in range(num_batches):

            end = start + batchs_size

            preds_to_test = prediction_shuffled[start:end]
            labels_to_test = label_shuffled[start:end]

            batches.append(hsic(preds_to_test, labels_to_test, 0.7, 0.5))

            start += batchs_size

        return np.mean(np.array(batches)).item()

    @property
    def name(self) -> str:
        return "HSIC"

    @property
    def apply_per_sensitive(self) -> bool:
        return False



"""Implementation of SVM (actually just a wrapper around sklearn)"""
from sklearn.svm import SVC, LinearSVC

import pandas as pd

from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.implementations.utils import InAlgoInterface


def select_svm(C: float, kernel: str):
    """Select the appropriate SVM model for the given parameters"""
    if kernel == "linear":
        return LinearSVC(C=C, dual=False, tol=1e-12, random_state=888)
    return SVC(C=C, kernel=kernel, gamma="auto", random_state=888)


def train_and_predict(train: DataTuple, test: TestTuple, C: float, kernel: str) -> pd.DataFrame:
    """Train an SVM model and compute predictions on the given test data"""
    clf = select_svm(C, kernel)
    clf.fit(train.x, train.y.to_numpy().ravel())
    return pd.DataFrame(clf.predict(test.x), columns=["preds"])


def main():
    """This function runs the SVM model as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    C, kernel = interface.remaining_args()
    interface.save_predictions(train_and_predict(train, test, C=float(C), kernel=kernel))


if __name__ == "__main__":
    main()



"""
Split into train and test data
"""
from typing import Tuple
import numpy as np
from pandas.testing import assert_index_equal
import pandas as pd

from ethicml.utility.data_structures import DataTuple


def call_numpy_to_split(
    dataframe: pd.DataFrame, train_percentage, random_seed
) -> (Tuple[pd.DataFrame, pd.DataFrame]):
    """Split a DataFrame along the rows"""
    # permute
    dataframe = dataframe.sample(frac=1, random_state=random_seed).reset_index(drop=True)

    # split
    train_len = int(train_percentage * len(dataframe))
    train = dataframe.iloc[:train_len]
    test = dataframe.iloc[train_len:]

    assert isinstance(train, pd.DataFrame)
    assert isinstance(test, pd.DataFrame)

    train = train.reset_index(drop=True)
    test = test.reset_index(drop=True)

    return train, test


def train_test_split(
    data: DataTuple, train_percentage: float = 0.8, random_seed: int = 0
) -> (Tuple[DataTuple, DataTuple]):
    """Split a data tuple into two datatuple along the rows of the DataFrames"""
    x_columns: pd.Index = data.x.columns
    s_columns: pd.Index = data.s.columns
    y_columns: pd.Index = data.y.columns

    all_data: pd.DataFrame = pd.concat([data.x, data.s, data.y], axis="columns")

    all_data = all_data.sample(frac=1, random_state=1).reset_index(drop=True)

    all_data_train_test: Tuple[pd.DataFrame, pd.DataFrame] = call_numpy_to_split(
        all_data, train_percentage, random_seed=random_seed
    )

    all_data_train, all_data_test = all_data_train_test

    train: DataTuple = DataTuple(
        x=all_data_train[x_columns],
        s=all_data_train[s_columns],
        y=all_data_train[y_columns],
        name=f"{data.name} - Train",
    )

    test: DataTuple = DataTuple(
        x=all_data_test[x_columns],
        s=all_data_test[s_columns],
        y=all_data_test[y_columns],
        name=f"{data.name} - Test",
    )

    assert isinstance(train.x, pd.DataFrame)
    assert isinstance(test.x, pd.DataFrame)
    assert_index_equal(train.x.columns, x_columns)
    assert_index_equal(test.x.columns, x_columns)

    assert isinstance(train.s, pd.DataFrame)
    assert isinstance(test.s, pd.DataFrame)
    assert_index_equal(train.s.columns, s_columns)
    assert_index_equal(test.s.columns, s_columns)

    assert isinstance(train.y, pd.DataFrame)
    assert isinstance(test.y, pd.DataFrame)
    assert_index_equal(train.y.columns, y_columns)
    assert_index_equal(test.y.columns, y_columns)

    return train, test


def fold_data(data: DataTuple, folds: int):
    """
    So much love to sklearn for making their source code open
    """

    indices: np.ndarray[int] = np.arange(data.x.shape[0])

    fold_sizes: np.ndarray[int] = np.full(folds, data.x.shape[0] // folds, dtype=int)
    fold_sizes[: data.x.shape[0] % folds] += 1

    current = 0
    for i, fold_size in enumerate(fold_sizes):
        start, stop = current, current + fold_size
        val_inds: np.ndarray[int] = indices[start:stop]
        train_inds = [i for i in indices if i not in val_inds]  # Pretty sure this is inefficient

        train_x = data.x.iloc[train_inds].reset_index(drop=True)
        train_s = data.s.iloc[train_inds].reset_index(drop=True)
        train_y = data.y.iloc[train_inds].reset_index(drop=True)

        assert train_x.shape == (len(train_inds), data.x.shape[1])
        assert train_s.shape == (len(train_inds), data.s.shape[1])
        assert train_y.shape == (len(train_inds), data.y.shape[1])

        val_x = data.x.iloc[val_inds].reset_index(drop=True)
        val_s = data.s.iloc[val_inds].reset_index(drop=True)
        val_y = data.y.iloc[val_inds].reset_index(drop=True)

        assert val_x.shape == (len(val_inds), data.x.shape[1])
        assert val_s.shape == (len(val_inds), data.s.shape[1])
        assert val_y.shape == (len(val_inds), data.y.shape[1])

        yield DataTuple(
            x=train_x, s=train_s, y=train_y, name=f"{data.name} - train fold {i}"
        ), DataTuple(x=val_x, s=val_s, y=val_y, name=f"{data.name} - test fold {i}")

        current = stop



"""
This is a kind of complicated model, but it's incredibly useful.
Say you find a papaer form a few years ago with code. It's not unreasonable that there might
be dependency clashes, python clashes, clashes galore. This approach downloads a model, runs it
in it's own venv and makes everyone happy.
"""

import os
from pathlib import Path
import shutil
import subprocess

import git

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.algorithms.inprocess.interface import conventional_interface
from ethicml.common import ROOT_PATH


ROOT_DIR = ROOT_PATH.parent


class InstalledModel(InAlgorithmAsync):
    """ the model that does the magic"""

    def __init__(self, name: str, url: str, module: str, file_name: str):
        self.repo_name = name
        self.module = module
        self.file_name = file_name
        self.url = url
        self.clone_directory()
        self.create_venv()
        super().__init__()

    @property
    def _executable(self) -> str:
        return str(self._module_path() / ".venv" / "bin" / "python")

    @property
    def name(self) -> str:
        return self.module

    def clone_directory(self):
        """
        Clones the repo
        """
        directory = Path(".") / self.repo_name
        if not os.path.exists(directory):
            os.makedirs(directory)
            git.Git(directory).clone(self.url)

    def create_venv(self):
        """
        Creates a venv based on the repos Pipfile
        """
        environ = os.environ.copy()
        environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
        environ["PIPENV_VENV_IN_PROJECT"] = "true"
        environ["PIPENV_YES"] = "true"
        environ["PIPENV_PIPFILE"] = str(self._module_path() / "Pipfile")

        venv_directory = self._module_path() / ".venv"

        if not os.path.exists(venv_directory):
            subprocess.check_call("pipenv install", env=environ, shell=True)

    def _module_path(self):
        return Path(".") / self.repo_name / self.module

    def _script_command(self, train_paths, test_paths, pred_path):
        """
        Overridden from parent - see there
        """
        args = conventional_interface(train_paths, test_paths, pred_path)
        return [str(self._module_path() / self.file_name)] + args

    def remove(self):
        """
        Removes the directory that we created in clone_directory
        """
        directory = Path(".") / self.repo_name
        try:
            shutil.rmtree(directory)
        except OSError as excep:
            print("Error: %s - %s." % (excep.filename, excep.strerror))



"""
Abstract Base Class of all metrics in the framework
"""

from abc import ABC, abstractmethod

import pandas as pd

from ethicml.utility.data_structures import DataTuple


class Metric(ABC):
    """Base class for all metrics"""

    def __init__(self, pos_class: int = 1):
        self.positive_class = pos_class

    @abstractmethod
    def score(self, prediction: pd.DataFrame, actual: DataTuple) -> float:
        """
        Compute score

        Args:
            prediction: predicted labels
            actual: dictionary with the actual labels and the sensitive attributes

        Returns:
            the score as a single number
        """

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the metric"""
        raise NotImplementedError()

    @property
    def apply_per_sensitive(self) -> bool:
        """
        Whether the metric can be applied per sensitive attribute
        """
        return True



"""Implementation of MLP (actually just a wrapper around sklearn)"""
from typing import Tuple
from ast import literal_eval as make_tuple

from sklearn.neural_network import MLPClassifier

import pandas as pd

from ethicml.implementations.utils import InAlgoInterface


def select_mlp(hidden_layer_sizes: Tuple[int], activation: str):
    """Create MLP model for the given parameters"""

    assert activation in ["identity", "logistic", "tanh", "relu"]

    return MLPClassifier(
        hidden_layer_sizes=hidden_layer_sizes, activation=activation, random_state=888
    )


def train_and_predict(train, test, hid_layers, activation):
    """Train an SVM model and compute predictions on the given test data"""
    clf = select_mlp(hid_layers, activation)
    clf.fit(train.x, train.y.values.ravel())
    return pd.DataFrame(clf.predict(test.x), columns=["preds"])


def main():
    """This function runs the SVM model as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    hid_layers, activation = interface.remaining_args()
    interface.save_predictions(
        train_and_predict(train, test, hid_layers=make_tuple(hid_layers), activation=activation)
    )


if __name__ == "__main__":
    main()



"""Implementation of logistic regression (actually just a wrapper around sklearn)"""
import pandas as pd

from fairlearn.classred import expgrad
from fairlearn.moments import Moment, DP, EO
from sklearn.linear_model import LogisticRegression

from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.implementations.utils import InAlgoInterface
from ethicml.implementations.svm import select_svm
from ethicml.utility.heaviside import Heaviside


def train_and_predict(
    train: DataTuple,
    test: TestTuple,
    classifier: str,
    fairness: str,
    eps: float,
    iters: int,
    C: float,
    kernel: str,
):
    """Train a logistic regression model and compute predictions on the given test data"""

    fairness_class: Moment = DP() if fairness == "DP" else EO()
    if classifier == "SVM":
        model = select_svm(C, kernel)
    else:
        model = LogisticRegression(solver="liblinear", random_state=888, max_iter=5000, C=float(C))

    data_x = train.x
    data_y = train.y[train.y.columns[0]]
    data_a = train.s[train.s.columns[0]]

    res_tuple = expgrad(
        dataX=data_x,
        dataA=data_a,
        dataY=data_y,
        learner=model,
        cons=fairness_class,
        eps=eps,
        T=iters,
    )

    res = res_tuple._asdict()

    preds = pd.DataFrame(res["best_classifier"](test.x), columns=["preds"])
    helper = Heaviside()
    preds = preds.apply(helper.apply)
    min_class_label = train.y[train.y.columns[0]].min()
    if preds["preds"].min() != preds["preds"].max():
        preds = preds.replace(preds["preds"].min(), min_class_label)
    return preds


def main():
    """This function runs the Agarwal model as a standalone program"""
    interface = InAlgoInterface()
    train, test = interface.load_data()
    classifier, fairness, eps, iters, C, kernel = interface.remaining_args()
    interface.save_predictions(
        train_and_predict(
            train,
            test,
            classifier=classifier,
            fairness=fairness,
            eps=float(eps),
            iters=int(iters),
            C=float(C),
            kernel=kernel,
        )
    )


if __name__ == "__main__":
    main()



"""
Applies sci-kit learn's confusion matrix
"""

from typing import Tuple
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix as conf_mtx

from ethicml.utility.data_structures import DataTuple


class LabelOutOfBounds(Exception):
    """Metric Not Applicable per sensitive attribute, apply to whole dataset instead"""


def confusion_matrix(
    prediction: pd.DataFrame, actual: DataTuple, pos_cls: int
) -> Tuple[int, int, int, int]:
    """Apply sci-kit learn's confusion matrix"""
    actual_y: np.ndarray[int] = actual.y.values
    labels: np.ndarray[int] = np.unique(actual_y)
    if labels.size == 1:
        labels = np.array([0, 1])
    conf_matr: np.ndarray = conf_mtx(y_true=actual_y, y_pred=prediction, labels=labels)

    if not pos_cls in labels:
        raise LabelOutOfBounds("Positive class specified must exist in the test set")

    tp_idx: int = (labels == pos_cls).nonzero()[0].item()

    true_pos = conf_matr[tp_idx, tp_idx]
    false_pos = conf_matr[:, tp_idx].sum() - true_pos
    false_neg = conf_matr[tp_idx, :].sum() - true_pos
    true_neg = conf_matr.sum() - true_pos - false_pos - false_neg
    return true_neg, false_pos, false_neg, true_pos



"""
Class to describe features of the SQF dataset
"""

from ethicml.data.dataset import Dataset


class Sqf(Dataset):
    """Stop, question and frisk dataset"""

    def __init__(self, split: str = "Sex", discrete_only: bool = False):
        super().__init__()
        self.discrete_only = discrete_only
        self.features = [
            "perstop",
            "sex",
            "race",
            "age",
            "ht_feet",
            "ht_inch",
            "weight",
            "weapon",
            "typeofid_O",
            "typeofid_P",
            "typeofid_R",
            "typeofid_V",
            "explnstp_N",
            "explnstp_Y",
            "othpers_N",
            "othpers_Y",
            "arstmade_N",
            "arstmade_Y",
            "offunif_N",
            "offunif_Y",
            "frisked_N",
            "frisked_Y",
            "searched_N",
            "searched_Y",
            "pf-hands_N",
            "pf-hands_Y",
            "pf-wall_N",
            "pf-wall_Y",
            "pf-grnd_N",
            "pf-grnd_Y",
            "pf-drwep_N",
            "pf-drwep_Y",
            "pf-ptwep_N",
            "pf-ptwep_Y",
            "pf-baton_N",
            "pf-baton_Y",
            "pf-hcuff_N",
            "pf-hcuff_Y",
            "pf-pepsp_N",
            "pf-pepsp_Y",
            "pf-other_N",
            "pf-other_Y",
            "radio_N",
            "radio_Y",
            "ac-rept_N",
            "ac-rept_Y",
            "ac-inves_N",
            "ac-inves_Y",
            "rf-vcrim_N",
            "rf-vcrim_Y",
            "rf-othsw_N",
            "rf-othsw_Y",
            "ac-proxm_N",
            "ac-proxm_Y",
            "rf-attir_N",
            "rf-attir_Y",
            "cs-objcs_N",
            "cs-objcs_Y",
            "cs-descr_N",
            "cs-descr_Y",
            "cs-casng_N",
            "cs-casng_Y",
            "cs-lkout_N",
            "cs-lkout_Y",
            "rf-vcact_N",
            "rf-vcact_Y",
            "cs-cloth_N",
            "cs-cloth_Y",
            "cs-drgtr_N",
            "cs-drgtr_Y",
            "ac-evasv_N",
            "ac-evasv_Y",
            "ac-assoc_N",
            "ac-assoc_Y",
            "cs-furtv_N",
            "cs-furtv_Y",
            "rf-rfcmp_N",
            "rf-rfcmp_Y",
            "ac-cgdir_N",
            "ac-cgdir_Y",
            "rf-verbl_N",
            "rf-verbl_Y",
            "cs-vcrim_N",
            "cs-vcrim_Y",
            "cs-bulge_N",
            "cs-bulge_Y",
            "cs-other_N",
            "cs-other_Y",
            "ac-incid_N",
            "ac-incid_Y",
            "ac-time_N",
            "ac-time_Y",
            "rf-knowl_N",
            "rf-knowl_Y",
            "ac-stsnd_N",
            "ac-stsnd_Y",
            "ac-other_N",
            "ac-other_Y",
            "sb-hdobj_N",
            "sb-hdobj_Y",
            "sb-outln_N",
            "sb-outln_Y",
            "sb-admis_N",
            "sb-admis_Y",
            "sb-other_N",
            "sb-other_Y",
            "rf-furt_N",
            "rf-furt_Y",
            "rf-bulg_N",
            "rf-bulg_Y",
            "offverb_ ",
            "offverb_V",
            "offshld_ ",
            "offshld_S",
            "forceuse_ ",
            "forceuse_DO",
            "forceuse_DS",
            "forceuse_OR",
            "forceuse_OT",
            "forceuse_SF",
            "forceuse_SW",
            "haircolr_BA",
            "haircolr_BK",
            "haircolr_BL",
            "haircolr_BR",
            "haircolr_DY",
            "haircolr_GY",
            "haircolr_RA",
            "haircolr_SN",
            "haircolr_SP",
            "haircolr_XX",
            "haircolr_ZZ",
            "eyecolor_BK",
            "eyecolor_BL",
            "eyecolor_BR",
            "eyecolor_DF",
            "eyecolor_GR",
            "eyecolor_GY",
            "eyecolor_HA",
            "eyecolor_XX",
            "eyecolor_Z",
            "build_H",
            "build_M",
            "build_T",
            "build_U",
            "build_Z",
        ]

        self.continuous_features = ["perstop", "ht_feet", "age", "ht_inch", "perobs", "weight"]

        if split == "Sex":
            self.sens_attrs = ["sex"]
            self.s_prefix = ["sex"]
            self.class_labels = ["weapon"]
            self.class_label_prefix = ["weapon"]
        elif split == "Race":
            self.sens_attrs = ["race"]
            self.s_prefix = ["race"]
            self.class_labels = ["weapon"]
            self.class_label_prefix = ["weapon"]
        elif split == "Race-Sex":
            self.sens_attrs = ["sex", "race"]
            self.s_prefix = ["race", "sex"]
            self.class_labels = ["weapon"]
            self.class_label_prefix = ["weapon"]
        else:
            raise NotImplementedError

    @property
    def name(self) -> str:
        return "SQF"

    @property
    def filename(self) -> str:
        return "sqf.csv"



"""
Test that we can get some metrics on predictions
"""

from typing import Tuple
import pandas as pd
import pytest
from pytest import approx

from ethicml.algorithms.inprocess import InAlgorithm, LRProb, SVM, LR, Kamiran
from ethicml.utility.data_structures import DataTuple
from ethicml.data import Adult, NonBinaryToy
from ethicml.data.load import load_data
from ethicml.evaluators.evaluate_models import run_metrics
from ethicml.evaluators.per_sensitive_attribute import (
    metric_per_sensitive_attribute, diff_per_sensitive_attribute,
    ratio_per_sensitive_attribute, MetricNotApplicable
)
from ethicml.metrics import (Accuracy, BCR, CV, Metric, NMI, PPV, NPV, ProbNeg,
                             ProbOutcome, ProbPos, TNR, TPR)
from ethicml.metrics.confusion_matrix import LabelOutOfBounds
from ethicml.metrics.hsic import Hsic
from ethicml.metrics.theil import Theil
from ethicml.preprocessing.train_test_split import train_test_split
from tests.run_algorithm_test import get_train_test

RTOL = 1e-5  # relative tolerance when comparing two floats


def test_get_acc_of_predictions():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc: Metric = Accuracy()
    assert acc.name == "Accuracy"
    score = acc.score(predictions, test)
    assert score == 0.89


def test_mni_preds_and_s():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc: Metric = NMI(base='s')
    assert acc.name == "NMI preds and s"
    score = acc.score(predictions, test)
    assert score == pytest.approx(0.083, abs=0.001)


def test_accuracy_per_sens_attr():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert acc_per_sens == {'s_0': 0.905, 's_1': 0.875}


def test_probpos_per_sens_attr():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbPos())
    assert acc_per_sens == {'s_0': 0.335, 's_1': 0.67}


def test_proboutcome_per_sens_attr():
    train, test = get_train_test()
    model: InAlgorithm = LRProb()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbOutcome())
    assert acc_per_sens == {'s_0': pytest.approx(0.372, abs=0.001),
                            's_1': pytest.approx(0.661, abs=0.001)}


def test_probneg_per_sens_attr():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbNeg())
    assert acc_per_sens == {'s_0': 0.665, 's_1': 0.33}


def test_acc_per_nonbinary_sens():
    data: DataTuple = load_data(Adult("Nationality"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())

    test_dict = {'native-country_Cambodia_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Cambodia_1':    pytest.approx(0.500, abs=0.001),
                 'native-country_Canada_0':      pytest.approx(0.779, abs=0.001),
                 'native-country_Canada_1':      pytest.approx(0.595, abs=0.001),
                 'native-country_China_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_China_1':       pytest.approx(0.800, abs=0.001),
                 'native-country_Columbia_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Columbia_1':    pytest.approx(1.000, abs=0.001),
                 'native-country_Cuba_0':        pytest.approx(0.778, abs=0.001),
                 'native-country_Cuba_1':        pytest.approx(0.826, abs=0.001),
                 'native-country_Dominican-Republic_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Dominican-Republic_1': pytest.approx(0.933, abs=0.001),
                 'native-country_Ecuador_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Ecuador_1':     pytest.approx(0.800, abs=0.001),
                 'native-country_El-Salvador_0': pytest.approx(0.778, abs=0.001),
                 'native-country_El-Salvador_1': pytest.approx(1.000, abs=0.001),
                 'native-country_England_0':     pytest.approx(0.779, abs=0.001),
                 'native-country_England_1':     pytest.approx(0.600, abs=0.001),
                 'native-country_France_0':      pytest.approx(0.778, abs=0.001),
                 'native-country_France_1':      pytest.approx(0.500, abs=0.001),
                 'native-country_Germany_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Germany_1':     pytest.approx(0.679, abs=0.001),
                 'native-country_Greece_0':      pytest.approx(0.778, abs=0.001),
                 'native-country_Greece_1':      pytest.approx(0.636, abs=0.001),
                 'native-country_Guatemala_0':   pytest.approx(0.778, abs=0.001),
                 'native-country_Guatemala_1':   pytest.approx(0.952, abs=0.001),
                 'native-country_Haiti_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_Haiti_1':       pytest.approx(0.929, abs=0.001),
                 'native-country_Holand-Netherlands_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Honduras_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Honduras_1':    pytest.approx(1.000, abs=0.001),
                 'native-country_Hong_0':        pytest.approx(0.778, abs=0.001),
                 'native-country_Hong_1':        pytest.approx(0.667, abs=0.001),
                 'native-country_Hungary_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Hungary_1':     pytest.approx(1.000, abs=0.001),
                 'native-country_India_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_India_1':       pytest.approx(0.589, abs=0.001),
                 'native-country_Iran_0':        pytest.approx(0.778, abs=0.001),
                 'native-country_Iran_1':        pytest.approx(0.364, abs=0.001),
                 'native-country_Ireland_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Ireland_1':     pytest.approx(0.818, abs=0.001),
                 'native-country_Italy_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_Italy_1':       pytest.approx(0.565, abs=0.001),
                 'native-country_Jamaica_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Jamaica_1':     pytest.approx(0.800, abs=0.001),
                 'native-country_Japan_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_Japan_1':       pytest.approx(0.760, abs=0.001),
                 'native-country_Laos_0':        pytest.approx(0.778, abs=0.001),
                 'native-country_Laos_1':        pytest.approx(1.000, abs=0.001),
                 'native-country_Mexico_0':      pytest.approx(0.774, abs=0.001),
                 'native-country_Mexico_1':      pytest.approx(0.963, abs=0.001),
                 'native-country_Nicaragua_0':   pytest.approx(0.778, abs=0.001),
                 'native-country_Nicaragua_1':   pytest.approx(0.875, abs=0.001),
                 'native-country_Outlying-US(Guam-USVI-etc)_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Outlying-US(Guam-USVI-etc)_1': pytest.approx(0.800, abs=0.001),
                 'native-country_Peru_0':        pytest.approx(0.778, abs=0.001),
                 'native-country_Peru_1':        pytest.approx(0.923, abs=0.001),
                 'native-country_Philippines_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Philippines_1': pytest.approx(0.717, abs=0.001),
                 'native-country_Poland_0':      pytest.approx(0.778, abs=0.001),
                 'native-country_Poland_1':      pytest.approx(0.813, abs=0.001),
                 'native-country_Portugal_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Portugal_1':    pytest.approx(0.800, abs=0.001),
                 'native-country_Puerto-Rico_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Puerto-Rico_1': pytest.approx(0.921, abs=0.001),
                 'native-country_Scotland_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Scotland_1':    pytest.approx(1.000, abs=0.001),
                 'native-country_South_0':       pytest.approx(0.778, abs=0.001),
                 'native-country_South_1':       pytest.approx(0.714, abs=0.001),
                 'native-country_Taiwan_0':      pytest.approx(0.778, abs=0.001),
                 'native-country_Taiwan_1':      pytest.approx(0.429, abs=0.001),
                 'native-country_Thailand_0':    pytest.approx(0.778, abs=0.001),
                 'native-country_Thailand_1':    pytest.approx(0.750, abs=0.001),
                 'native-country_Trinadad&Tobago_0': pytest.approx(0.778, abs=0.001),
                 'native-country_Trinadad&Tobago_1': pytest.approx(1.000, abs=0.001),
                 'native-country_United-States_0': pytest.approx(0.806, abs=0.001),
                 'native-country_United-States_1': pytest.approx(0.776, abs=0.001),
                 'native-country_Vietnam_0':     pytest.approx(0.778, abs=0.001),
                 'native-country_Vietnam_1':     pytest.approx(0.929, abs=0.001),
                 'native-country_Yugoslavia_0':  pytest.approx(0.778, abs=0.001),
                 'native-country_Yugoslavia_1':  pytest.approx(0.714, abs=0.001)}

    for k, v in acc_per_sens.items():
        assert acc_per_sens[k] == test_dict[k]


def test_acc_per_race():
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())
    test_dict = {'race_Amer-Indian-Eskimo_0': pytest.approx(0.78, abs=0.01),
                 'race_Amer-Indian-Eskimo_1': pytest.approx(0.94, abs=0.01),
                 'race_Asian-Pac-Islander_0': pytest.approx(0.78, abs=0.01),
                 'race_Asian-Pac-Islander_1': pytest.approx(0.73, abs=0.01),
                 'race_Black_0': pytest.approx(0.77, abs=0.01),
                 'race_Black_1': pytest.approx(0.87, abs=0.01),
                 'race_Other_0': pytest.approx(0.78, abs=0.01),
                 'race_Other_1': pytest.approx(0.88, abs=0.01),
                 'race_White_0': pytest.approx(0.85, abs=0.01),
                 'race_White_1': pytest.approx(0.77, abs=0.01)}

    for k, v in acc_per_sens.items():
        assert acc_per_sens[k] == test_dict[k]


def test_tpr_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    assert tprs == {'s_0': pytest.approx(0.84, abs=0.01), 's_1': pytest.approx(0.88, abs=0.01)}
    tpr_diff = diff_per_sensitive_attribute(tprs)
    print(tpr_diff)
    assert tpr_diff["s_0-s_1"] == pytest.approx(0.04, abs=0.01)


def test_get_nmi_of_predictions():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmi: Metric = NMI()
    assert nmi.name == "NMI preds and y"
    score = nmi.score(predictions, test)
    assert score == pytest.approx(0.50, abs=0.01)


def test_nmi_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, NMI())
    assert NMI().name == "NMI preds and y"
    assert nmis == {'s_0': pytest.approx(0.52, abs=0.01), 's_1': pytest.approx(0.42, abs=0.01)}
    nmi_diff = diff_per_sensitive_attribute(nmis)
    assert nmi_diff["s_0-s_1"] == pytest.approx(0.10, abs=0.01)


def test_ppv_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, PPV())
    assert PPV().name == "PPV"
    assert results == {'s_0': pytest.approx(0.88, abs=0.01), 's_1': pytest.approx(0.93, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.05, abs=0.1)


def test_npv_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, NPV())
    assert NPV().name == "NPV"
    assert results == {'s_0': pytest.approx(0.91, abs=0.01), 's_1': pytest.approx(0.75, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.15, abs=0.01)


def test_bcr_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, BCR())
    assert BCR().name == "BCR"
    assert results == {'s_0': pytest.approx(0.89, abs=0.01), 's_1': pytest.approx(0.86, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.02, abs=0.01)


def test_cv():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    cv = CV()
    score = cv.score(predictions, test)
    assert CV().name == "CV"
    assert score == 0.665


def test_theil():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert Theil().name == "Theil_Index"
    assert score == pytest.approx(0.08, abs=0.01)

    model = LR()
    predictions = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert score == pytest.approx(0.07, abs=0.01)

    model = Kamiran()
    predictions = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert score == pytest.approx(0.08, abs=0.01)


def test_theil_per_sens_attr():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    theil_per_sens = metric_per_sensitive_attribute(predictions, test, Theil())
    assert theil_per_sens == {'s_0': pytest.approx(0.07, abs=0.01), 's_1': pytest.approx(0.10, abs=0.01)}


def test_hsic():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    hsic = Hsic()
    score = hsic.score(predictions, test)
    assert Hsic().name == "HSIC"
    assert score == pytest.approx(0.02, abs=0.01)


def test_use_appropriate_metric():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    with pytest.raises(MetricNotApplicable):
        metric_per_sensitive_attribute(predictions, test, CV())


def test_tnr_diff():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, TNR())
    assert NMI().name == "NMI preds and y"
    assert nmis == {'s_0': pytest.approx(0.93, abs=0.01), 's_1': pytest.approx(0.84, abs=0.01)}
    nmi_diff = diff_per_sensitive_attribute(nmis)
    assert nmi_diff["s_0-s_1"] == pytest.approx(0.09, abs=0.01)


def test_run_metrics():
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = run_metrics(predictions, test, [CV()], [TPR()])
    assert len(results) == 5
    assert results['TPR_s_0'] == approx(0.842857, RTOL)
    assert results['TPR_s_1'] == approx(0.886525, RTOL)
    assert results['TPR_s_0-s_1'] == approx(abs(0.842857 - 0.886525), RTOL)
    assert results['TPR_s_0/s_1'] == approx(0.842857 / 0.886525, RTOL)
    assert results['CV'] == approx(0.665)


def test_nmi_diff_non_binary_race():
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, NMI())
    assert NMI().name == "NMI preds and y"
    test_dict = {'race_Amer-Indian-Eskimo_0': pytest.approx(0.14, abs=0.01),
                 'race_Amer-Indian-Eskimo_1': pytest.approx(0.41, abs=0.01),
                 'race_Asian-Pac-Islander_0': pytest.approx(0.14, abs=0.01),
                 'race_Asian-Pac-Islander_1': pytest.approx(0.10, abs=0.01),
                 'race_Black_0': pytest.approx(0.14, abs=0.01),
                 'race_Black_1': pytest.approx(0.15, abs=0.01),
                 'race_Other_0': pytest.approx(0.14, abs=0.01),
                 'race_Other_1': pytest.approx(0.18, abs=0.01),
                 'race_White_0': pytest.approx(0.16, abs=0.01),
                 'race_White_1': pytest.approx(0.14, abs=0.01)}

    for k, v in nmis.items():
        assert nmis[k] == test_dict[k]

    nmis_to_check = {k: nmis[k] for k in ('race_Amer-Indian-Eskimo_1',
                                          'race_Asian-Pac-Islander_1',
                                          'race_Black_1',
                                          'race_Other_1',
                                          'race_White_1')}
    nmi_diff = diff_per_sensitive_attribute(nmis_to_check)
    test_dict = {'race_Amer-Indian-Eskimo_1-race_Asian-Pac-Islander_1': pytest.approx(0.31, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_Black_1': pytest.approx(0.26, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_Other_1': pytest.approx(0.23, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_White_1': pytest.approx(0.27, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_Black_1': pytest.approx(0.04, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_Other_1': pytest.approx(0.08, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_White_1': pytest.approx(0.03, abs=0.01),
                 'race_Black_1-race_Other_1': pytest.approx(0.03, abs=0.01),
                 'race_Black_1-race_White_1': pytest.approx(0.02, abs=0.01),
                 'race_Other_1-race_White_1': pytest.approx(0.04, abs=0.01)}

    for k, v in nmi_diff.items():
        assert nmi_diff[k] == test_dict[k]


def test_tpr_diff_non_binary_race():
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    test_dict = {'race_Amer-Indian-Eskimo_0': pytest.approx(0.16, abs=0.01),
                 'race_Amer-Indian-Eskimo_1': pytest.approx(0.37, abs=0.01),
                 'race_Asian-Pac-Islander_0': pytest.approx(0.16, abs=0.01),
                 'race_Asian-Pac-Islander_1': pytest.approx(0.12, abs=0.01),
                 'race_Black_0': pytest.approx(0.16, abs=0.01),
                 'race_Black_1': pytest.approx(0.13, abs=0.01),
                 'race_Other_0': pytest.approx(0.16, abs=0.01),
                 'race_Other_1': pytest.approx(0.12, abs=0.01),
                 'race_White_0': pytest.approx(0.14, abs=0.01),
                 'race_White_1': pytest.approx(0.16, abs=0.01)}

    for k, v in tprs.items():
        assert tprs[k] == test_dict[k]

    tprs_to_check = {k: tprs[k] for k in ('race_Amer-Indian-Eskimo_1',
                                          'race_Asian-Pac-Islander_1',
                                          'race_Black_1',
                                          'race_Other_1',
                                          'race_White_1')}
    tpr_diff = diff_per_sensitive_attribute(tprs_to_check)
    test_dict = {'race_Amer-Indian-Eskimo_1-race_Asian-Pac-Islander_1': pytest.approx(0.25, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_Black_1': pytest.approx(0.23, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_Other_1': pytest.approx(0.25, abs=0.01),
                 'race_Amer-Indian-Eskimo_1-race_White_1': pytest.approx(0.20, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_Black_1': pytest.approx(0.01, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_Other_1': pytest.approx(0.00, abs=0.01),
                 'race_Asian-Pac-Islander_1-race_White_1': pytest.approx(0.04, abs=0.01),
                 'race_Black_1-race_Other_1': pytest.approx(0.01, abs=0.01),
                 'race_Black_1-race_White_1': pytest.approx(0.04, abs=0.01),
                 'race_Other_1-race_White_1': pytest.approx(0.04, abs=0.01)}

    for k, v in tpr_diff.items():
        assert tpr_diff[k] == test_dict[k]


def test_tpr_ratio_non_binary_race():
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    test_dict = {'race_Amer-Indian-Eskimo_0': pytest.approx(0.16, abs=0.01),
                 'race_Amer-Indian-Eskimo_1': pytest.approx(0.37, abs=0.01),
                 'race_Asian-Pac-Islander_0': pytest.approx(0.16, abs=0.01),
                 'race_Asian-Pac-Islander_1': pytest.approx(0.12, abs=0.01),
                 'race_Black_0': pytest.approx(0.16, abs=0.01),
                 'race_Black_1': pytest.approx(0.13, abs=0.01),
                 'race_Other_0': pytest.approx(0.16, abs=0.01),
                 'race_Other_1': pytest.approx(0.12, abs=0.01),
                 'race_White_0': pytest.approx(0.14, abs=0.01),
                 'race_White_1': pytest.approx(0.16, abs=0.01)}

    for k, v in tprs.items():
        assert tprs[k] == test_dict[k]

    tprs_to_check = {k: tprs[k] for k in ('race_Amer-Indian-Eskimo_1',
                                          'race_Asian-Pac-Islander_1',
                                          'race_Black_1',
                                          'race_Other_1',
                                          'race_White_1')}
    tpr_diff = ratio_per_sensitive_attribute(tprs_to_check)
    test_dict = {'race_Amer-Indian-Eskimo_1/race_Asian-Pac-Islander_1': pytest.approx(0.32, abs=0.1),
                 'race_Amer-Indian-Eskimo_1/race_Black_1': pytest.approx(0.37, abs=0.1),
                 'race_Amer-Indian-Eskimo_1/race_Other_1': pytest.approx(0.33, abs=0.1),
                 'race_Amer-Indian-Eskimo_1/race_White_1': pytest.approx(0.44, abs=0.1),
                 'race_Asian-Pac-Islander_1/race_Black_1': pytest.approx(0.88, abs=0.1),
                 'race_Asian-Pac-Islander_1/race_Other_1': pytest.approx(0.97, abs=0.1),
                 'race_Asian-Pac-Islander_1/race_White_1': pytest.approx(0.72, abs=0.1),
                 'race_Black_1/race_Other_1': pytest.approx(0.91, abs=0.1),
                 'race_Black_1/race_White_1': pytest.approx(0.74, abs=0.1),
                 'race_Other_1/race_White_1': pytest.approx(0.74, abs=0.1)}

    for k, v in tpr_diff.items():
        assert tpr_diff[k] == test_dict[k]


def test_nb_acc():
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_score = Accuracy().score(predictions, test)
    assert acc_score == 0.1
    accs = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert accs == {'sens_0': pytest.approx(0.09, abs=0.1), 'sens_1': pytest.approx(0.11, abs=0.1)}
    model = LR()
    predictions = model.run_test(train, test)
    acc_score = Accuracy().score(predictions, test)
    assert acc_score == 0.7
    accs = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert accs == {'sens_0': pytest.approx(0.72, abs=0.1), 'sens_1': pytest.approx(0.66, abs=0.1)}


def test_nb_tpr():
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tpr_score = TPR(pos_class=1).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=2).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=3).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=4).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=5).score(predictions, test)
    assert tpr_score == 0.0

    with pytest.raises(LabelOutOfBounds):
        tpr_score = TPR(pos_class=0).score(predictions, test)

    accs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert accs == {'sens_0': pytest.approx(0.0, abs=0.1), 'sens_1': pytest.approx(0.0, abs=0.1)}

    model = LR()
    predictions = model.run_test(train, test)

    print(list([(k, z) for k, z in zip(predictions.values, test.y.values) if k != z]))

    tpr_score = TPR(pos_class=1).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=2).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=3).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=4).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=5).score(predictions, test)
    assert tpr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tpr_score = TPR(pos_class=0).score(predictions, test)

    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert tprs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}


def test_nb_tnr():
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tnr_score = TNR(pos_class=1).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=2).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=3).score(predictions, test)
    assert tnr_score == 0.0
    tnr_score = TNR(pos_class=4).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=5).score(predictions, test)
    assert tnr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tnr_score = TNR(pos_class=0).score(predictions, test)

    accs = metric_per_sensitive_attribute(predictions, test, TNR())
    assert accs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}

    model = LR()
    predictions = model.run_test(train, test)

    print(list([(k, z) for k, z in zip(predictions.values, test.y.values) if k != z]))

    tnr_score = TNR(pos_class=1).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=2).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=3).score(predictions, test)
    assert tnr_score == pytest.approx(0.7, abs=0.1)
    tnr_score = TNR(pos_class=4).score(predictions, test)
    assert tnr_score == pytest.approx(0.85, abs=0.1)
    tnr_score = TNR(pos_class=5).score(predictions, test)
    assert tnr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tnr_score = TNR(pos_class=0).score(predictions, test)

    tnrs = metric_per_sensitive_attribute(predictions, test, TNR())
    assert tnrs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}



"""
Abstract Base Class of all algorithms in the framework
"""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, List
from abc import abstractmethod

from ethicml.algorithms.algorithm_base import Algorithm, AlgorithmAsync, run_blocking
from ethicml.utility.data_structures import (
    get_subset,
    DataTuple,
    TestTuple,
    PathTuple,
    TestPathTuple,
    write_as_feather,
    load_feather,
)


class PreAlgorithm(Algorithm):
    """Abstract Base Class for all algorithms that do pre-processing"""

    @abstractmethod
    def run(self, train: DataTuple, test: TestTuple) -> (Tuple[DataTuple, TestTuple]):
        """Generate fair features with the given data

        Args:
            train: training data
            test: test data
        """

    def run_test(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Run with reduced training set so that it finishes quicker"""
        train_testing = get_subset(train)
        return self.run(train_testing, test)


class PreAlgorithmAsync(PreAlgorithm, AlgorithmAsync):
    """Pre-Algorithm that can be run blocking and asynchronously"""

    def run(self, train, test):
        return run_blocking(self.run_async(train, test))

    async def run_async(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Generate fair features with the given data asynchronously"""
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            train_paths, test_paths = write_as_feather(train, test, tmp_path)
            train_x_path = tmp_path / "transform_train_x.feather"
            train_s_path = tmp_path / "transform_train_s.feather"
            train_y_path = tmp_path / "transform_train_y.feather"
            train_name_path = tmp_path / "transform_train_name.feather"
            test_x_path = tmp_path / "transform_test_x.feather"
            test_s_path = tmp_path / "transform_test_s.feather"
            test_name_path = tmp_path / "transform_test_name.feather"
            cmd = self._script_command(
                train_paths,
                test_paths,
                train_x_path,
                train_s_path,
                train_y_path,
                train_name_path,
                test_x_path,
                test_s_path,
                test_name_path,
            )
            await self._call_script(cmd)
            return (
                DataTuple(
                    x=load_feather(train_x_path),
                    s=load_feather(train_s_path),
                    y=load_feather(train_y_path),
                    name=str(load_feather(train_name_path)['0'][0]),
                ),
                TestTuple(
                    x=load_feather(test_x_path),
                    s=load_feather(test_s_path),
                    name=str(load_feather(test_name_path)['0'][0]),
                ),
            )

    @abstractmethod
    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_x_path: Path,
        new_train_s_path: Path,
        new_train_y_path: Path,
        new_train_name_path: Path,
        new_test_x_path: Path,
        new_test_s_path: Path,
        new_test_name_path: Path,
    ) -> List[str]:
        """The command that will run the script"""
